"""
This file is part of the Hatkit Datafiddler, a tool for analyzing
MongoDB database content, particularly http traffic.
  
Copyright (C) 2011 Martin Holst Swende (hatkit@swende.se)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
'''
Created on Apr 5, 2010

@author: Martin Holst Swende
'''
import thirdparty
import subprocess
from subprocess import PIPE
from threading import Thread
import time
import sys
import socket
import httplib2
from httplib import HTTPConnection
import BaseHTTPServer

RATPROXY_PATH = "/home/martin/tools/ratproxy-src/"
RATPROXY_BIN = "/home/martin/tools/ratproxy-src/ratproxy"

# Ratproxy usage : 
#Usage: ./ratproxy [ -w logfile ] [ -v logdir ] [ -p port ] [ -d domain ] [ -P host:port ] [ -xtifkgmjscael2XCr ]
#   -w logfile    - write results to a specified file (default: stdout)
#   -v logdir     - write HTTP traces to a specified directory (default: none)
#   -p port       - listen on a custom TCP port (default: 8080)
#   -d domain     - analyze requests to specified domains only (default: all)
#   -P host:port  - use upstream proxy for all requests (format host:port)
#   -r            - accept remote connections (default: 127.0.0.1 only)
#   -l            - use response length, not checksum, for identity check
#   -2            - perform two, not one, page identity check
#   -e            - perform pedantic caching headers checks
#   -x            - log all XSS candidates
#   -t            - log all directory traversal candidates
#   -i            - log all PNG files served inline
#   -f            - log all Flash applications for analysis (add -v to decompile)
#   -s            - log all POST requests for analysis
#   -c            - log all cookie setting URLs for analysis
#   -g            - perform XSRF token checks on all GET requests
#   -j            - report on risky Javascript constructions
#   -m            - log all active content referenced across domains
#   -X            - disruptively validate XSRF, XSS protections
#   -C            - try to auto-correct persistent side effects of -X
#   -k            - flag HTTP requests as bad (for HTTPS-only applications)
#   -a            - indiscriminately report all visited URLs


class RatAnalyser(thirdparty.Analyser):
    def __init__(self):
        thirdparty.Analyser.__init__(self)

    def setup(self, ratPort=2727, localPort=2728):
        """Starts the ratproxy process and opens the local port as upstream proxy.
        Also starts two threads, one which reads 
        stdout and one which reads stderr from the ratproxy process 
        """
        # Save some variables
        self.ratPort = ratPort
        self.serverPort = localPort
        #Start the process
        args = [RATPROXY_BIN, '-p', '%d' % self.ratPort, '-P', 'localhost:%d' % self.serverPort, '-xtifscgjm']
        p = subprocess.Popen(args,stdout=PIPE,stderr=PIPE,close_fds=True , cwd=RATPROXY_PATH) 
        self.ratProcess = p;
        print("Process started as pid %d" % p.pid)
        
        # Hook up reader-threads for stderr and stdout
        self.ratreader_std = self.startReader(p.stdout,"stdout: ")
        self.ratreader_err = self.startReader(p.stderr,"stderr: ")
        self.ratreader_std.start()
        self.ratreader_err.start()

        # Open up a port which ratproxy can use as upstream proxy
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("localhost", self.serverPort))
        self.serversocket = s

    def shutDown(self):
        print "Killing ratproxy"
        self.ratProcess.kill()
        
        
    def startReader(self, fileobject, desc=""):
        """
        This method starts a thread which reads from 
        a file-like object
        """
        class RatReader(Thread):
            def run(self):
                self._exit = False
                #for l in fileobject:
                #    print desc+l
                while not self._exit:
                    c= fileobject.readline()
                    if(c == "") :
                        break
                    sys.stdout.write( desc+c)
                    #if c != "":
                    #else:
                    #    time.sleep(1)
                print "Quitting %s " % desc
            def exit(self):
                self._exit = True
        return RatReader()
    def tearDown(self):
        self.ratreader.exit()
        self.ratProcess.terminate()
    def notNone(self, o):
        if o == None: return ""
        return o          
    
    def concatenateToString(self, array):
        p = [];
        keys = array.keys();
        for key in keys:
            p.append("%s: %s" % (key, array[key]))
        return "\n".join(p)
    
    def analyse_objects(self, req,res, rawreq, rawres):
        """
        This method takes request and response from the database, and sends the 
        request to the ratproxy. Ratproxy will analyse the request and forward
        it to the next upstream proxy (us). We read the incoming data from ratproxy
        and respond with the response from database, to let ratproxy analyse it aswell. 
        """
        rq_hdr = self.notNone(req["headers"])
        rq_body = self.notNone(req["data"])
        rq_start = req["startline"]
        ssl = req["ssl"]
        inet = req["target"]
        
        #Target may be on the form 
        #www.langbrovardshus.se/91.207.158.130:80
        if len(inet.split("/")) == 2 :
            inet = inet.split("/")[1]
        
        print rq_start
        #Now, we must replace
        # GET /foo.asp HTTP/1.1 => GET http://server:port/foo.asp HTTP/1.1
        req_split = rq_start.split(" ")
        if ssl:
            resource = "https://";
        else:
            resource = "http://";

        startline = req_split[0]+" "+ resource + inet+ req_split[1] +" "+ req_split[2]
        print startline;
        
        full_request = startline +"\r\n"+ self.concatenateToString(rq_hdr)+"\r\n\r\n"+self.concatenateToString(rq_body)
        
        
        res_hdrs = self.notNone(rawres["headers"])
        res_body = self.notNone(rawres["content"])
        res = res_hdrs + res_body


        # Connect to rat 
        rat = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        rat.connect(("localhost", self.ratPort))
        
        # do http proxy CONNECT
        #rat.send("CONNECT %s:%d HTTP/1.0\r\n\r\n" % ("rats.thats.me", 80))
        # Request sent to Ratproxy. Read request on our server port
        
        # Send request
        print "-req:-"
        print full_request
        print "---"
        print "Sending %d bytes to rat" % len(full_request)
        rat.send(full_request)
        
        self.serversocket.listen(1)
        conn, addr = self.serversocket.accept()
        print 'Rats dialled in: ', addr
        rat_in = conn.recv(1024)
        print "Rats says: %s" % rat_in
        #conn.send('HTTP/1.0 200 Gimme your best!\r\n')
        #conn.send('Proxy-agent: RatHat 0.3\r\n')
        #conn.send('\r\n')
        
        #p = rat.recv(200)
        #print "Connect response : %s" % p
        
        #data = conn.recv(len(req) + 100)
        #print "Read %d bytes from rat" % len(data)
        #Now, send response
        print "To rat:------------__"
        print res
        print "------------------"
        conn.send(res)
        #And close
        conn.close()
        
        #Read response from rat
        print(" Hoping to read %d bytes from rat" % len(res))
        data = rat.recv(len(res)+100)
        print ("Read %d bytes from rat" % len(data))
        rat.close()
#
# PROXY HTTP REQUEST HANDLER AND SERVER
# the following classes provide a simple implementation of a proxy server
#
class ProxyHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    server_version = 'Snaprats/0.2'

    def _sock_connect_to(self, netloc, soc):
        return True

if __name__ == '__main__':
    rat = RatAnalyser()
    try:
        rat.start()
    finally:
        rat.shutDown()
    #rat.setup()
#    rat.tearDown()
    #rat.start()
