"""
This file is part of the Hatkit Datafiddler, a tool for analyzing
MongoDB database content, particularly http traffic.
  
Copyright (C) 2011 Martin Holst Swende (hatkit@swende.se)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
'''
Created on Feb 11, 2011

@author: martin
'''
from PyQt4.Qt import *
from PyQt4 import QtGui, QtCore

from transforming.transform import transformer
from ui_generated import tabledata_settings, database_filter_widget
from ui.WindowHandling import WindowUtils
from ui.DatabaseFilterHandling import DatabaseFilterHandler
from ui.UIHelpers import ComboboxSettingsHandler

class TableviewSettingsWindow(tabledata_settings.Ui_Form):
    
    def __init__(self, mongoTablemodel, mongoTableView, prefs):
        self.mongoTableModel = mongoTablemodel
        self.mongoTableView = mongoTableView

    def setupUi(self, DBDialog):
        super(TableviewSettingsWindow, self).setupUi(DBDialog)
        
        #Add the database filter tab
        self.tab_filter = QtGui.QWidget()
        self.filter = filter = database_filter_widget.Ui_database_filter()
        filter.setupUi(self.tab_filter)
        self.tab_filter.setObjectName("tab_filter")
        self.tabWidget.addTab(self.tab_filter, "Database filtering")
        # The database filter tab has one filterhandler
        self.filterhandler = DatabaseFilterHandler(filter,self.mongoTableModel.query)

        #Configure the handler for columns/varialbes
        self.columnTableHandler = SimpleTableHandler(self.tableWidget_cols, self.mongoTableModel.getColumnDefinitions, self.mongoTableModel.setTransforms, ColumnDisplayDefinition)
        self.variableTableHandler = SimpleTableHandler(self.tableWidget_vars, self.mongoTableModel.getVariables, self.mongoTableModel.setVariables, DbVariableDefinition)
        self.toolButton_addvar.clicked.connect(self.variableTableHandler.addRow)
        self.toolButton_addcol.clicked.connect(self.columnTableHandler.addRow)
        self.pushButton_apply_vars_cols.clicked.connect(self.apply)
        self.pushButton_revert_vars_cols.clicked.connect(self.revert)
        self.pushButton_help_vars_cols.clicked.connect(self.showHelp)
        # Also add the  handler for the save/load using the "preferences"-module
        self.prefsSaver = ComboboxSettingsHandler(self.comboBoxLoad, self.pushButton_load_varscols, self.pushButtonSave, self.lineEditSaveName, self, "varsandcols-settings")
        
        
    def apply(self):
        #Modify the model
        self.variableTableHandler.apply()
        self.columnTableHandler.apply()
        #Fetch new data
        self.mongoTableModel.reload()
        self.mongoTableView.resizeColumnsToContents()
        
    def revert(self):
        self.variableTableHandler.revert()
        self.columnTableHandler.revert()
        
    def showHelp(self):
        html = transformer.getHelp()
        WindowUtils.showHelp(html)

    def getSaveData(self, name):
        """Returns the data which should be persisted if the user presses 'save' the current setup"""
        vars = self.variableTableHandler.getData()
        cols = self.columnTableHandler.getData()
        return {"vars":vars, "cols": cols}
    
    def applySaveData(self, data,name):
        """Applies whatever was saved in getSaveData"""
        self.variableTableHandler.setData(data['vars'])
        self.columnTableHandler.setData(data['cols'])


        
class SimpleTableHandler():
    """
    Class to handle the small tables which displays/edits variable loading and columns from database. 
    Uses a tablewidget
    """
    def __init__(self, table, getter, setter, factory, parent=None, *args): 
        """
        @param table: the tablewidget to handle
        @param getter: a function which can be called upon to get the current values to edit 
        @param setter: a function which can be called upon to set new values when they have been changed
        @param createNewObjectFunction: a function which can be called upon to create a new value-object
        """
        self.getter = getter
        self.setter = setter
        self.table = table
        self.factory = factory
        self.revert()
 
    def revert(self):
        """Call the 'getter' to revert and discard changes made by the user"""
        self.content = self.getter()
        self.initTable()
        self.table.resizeColumnsToContents()

    def initTable(self):
        """Clear and repopulate the table"""
        
        t = self.table
        t.clear()
        
        self.content[0].initTable(t);
        
        rownames = []
        
        t.setRowCount(len(self.content))        
        for x in range(len(self.content)):
            columndef = self.content[x]
            columndef.addToTable(t, x)
            rownames.append(columndef.getRowname(x))
        
        t.setVerticalHeaderLabels(rownames)
        
    def apply(self):
        """Apply changes made in the UI"""
        # Get the changed data
        self.getData()
        # Call the supplied setter-function
        self.setter(self.content)

    def getData(self):
        """Take whatever changes has been performed in the UI and
        internalizes that to the internal format (basically parses it)"""
        
        #Important here is to create *new* objects, not just 
        # parse into the old ones, since if the user saves
        # his preferences, we cannot overwrite those values
        newcontent = []
        for x in range(len(self.content)):
            newcontent.append(self.factory.parse(self.table, x))

        self.content = newcontent
        return self.content

    def addRow(self):
        """Adds a new row to the column"""
        self.getData()
        # Add a new item using the provided factory function
        self.content.append(self.factory.newObject())
        self.initTable()
        
    def setData(self, content):
        """This function is called when definitions are loaded from file
        It completely overwrites the 'current' data in both varialbes and columndefs"""
        self.content = content
        self.initTable()
        

class DbVariableDefinition():

    @staticmethod
    def initTable(t):
        cols = ['Variables']
        t.setColumnCount(len(cols)) 
        t.setHorizontalHeaderLabels(cols)

    @staticmethod
    def parse(table, index):
        """Interpret the UI form elements and create a new object based on it """
        expression = str(table.item(index, 0).text())
        return DbVariableDefinition(expression)
    
    @staticmethod
    def newObject():
        return DbVariableDefinition()
    
    @staticmethod
    def getRowname(index):
        return "v%s" % index
    
    def __init__(self, expression=""):
        # The expression for the variable to load
        self.expression = expression
    

    def addToTable(self, t, index):
        """ Add my fields to the table"""
        t.setItem(index, 0, QtGui.QTableWidgetItem(str(self.expression)))

        
class ColumnDisplayDefinition():
    """Class to help between transition between UI creation/editing and storage in the mongo table model"""
    
    @staticmethod
    def getRowname(index):
        return str(index)
    
    @staticmethod
    def initTable(t):
        cols = ['Column', 'Coloring', 'Enabled', 'Title']
        t.setColumnCount(len(cols)) 
        t.setHorizontalHeaderLabels(cols)

    @staticmethod
    def parse(table, index):
        """Interpret the UI form elements and create a new object """
        expression = str(table.item(index, 0).text())
        coloring = table.item(index, 1).checkState()
        enabled = (table.item(index, 2).checkState() == 2)
        title = str(table.item(index, 3).text())
        return ColumnDisplayDefinition(expression, coloring, enabled, title)

    @staticmethod
    def newObject():
        return ColumnDisplayDefinition()
    
    def __init__(self, expression="", coloring=1, enabled=True, title=None):
        
        # The expressio for the column transformation
        self.expression = expression
        # The coloring scheme for the column, 0,1 or 2
        self.coloring = coloring
        # If the column is active at all, or not used
        self.enabled = enabled
        #THe title of the column
        if title == None:
            self.title = self.expression
        else: 
            self.title = title
            
    def addToTable(self, t, index):
        coloring = QtGui.QTableWidgetItem("")
        coloring.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsTristate | QtCore.Qt.ItemIsEnabled)
        coloring.setCheckState(self.coloring)
        columnEnabled = QtGui.QTableWidgetItem("")
        columnEnabled.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
        if self.enabled:
            columnEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            columnEnabled.setCheckState(QtCore.Qt.Unchecked)


        t.setItem(index, 0, QtGui.QTableWidgetItem(str(self.expression)))
        t.setItem(index, 1, coloring)
        t.setItem(index, 2, columnEnabled)
        t.setItem(index, 3, QtGui.QTableWidgetItem(self.title))

