"""
This file is part of the Hatkit Datafiddler, a tool for analyzing
MongoDB database content, particularly http traffic.
  
Copyright (C) 2011 Martin Holst Swende (hatkit@swende.se)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
'''
Created on Mar 8, 2010

@author: Martin Holst Swende
'''
from PyQt4.Qt import *
from PyQt4.QtCore import * 
from PyQt4.QtGui import * 
import operator
import hatcore.DBConnector
from transforming.transform import transformer
from PyQt4 import QtGui
from ui.TabledataSettings import DbVariableDefinition, ColumnDisplayDefinition
from hatcore import DBQuerying

def getUniqueColor(input):
    bits = hash(input)
    r = (bits & 0xFF0000) >> 16
    g = (bits & 0x00FF00) >> 8
    b = (bits & 0x0000FF)
    return (r, g, b)

class QMongoTableModel(QAbstractTableModel):
    
    '''
    This is the data model which provides the table with database data. 
    
    When subclassing QAbstractTableModel, you must implement rowCount(), columnCount(), and data(). Default implementations of the index() and parent() functions are provided by QAbstractTableModel. Well behaved models will also implement headerData().
    http://qt.nokia.com/doc/4.0/qabstracttablemodel.html
    '''
    fields = [DbVariableDefinition("_id"),
              DbVariableDefinition("request.time"),
              DbVariableDefinition("request.headers"),
              DbVariableDefinition("request.url"),
              DbVariableDefinition("response.status"),
              DbVariableDefinition("response.headers") ]
    
    transforms = [ColumnDisplayDefinition("v0", Qt.Checked),
                  ColumnDisplayDefinition("date(v1)", Qt.Unchecked),
                  ColumnDisplayDefinition("v2['Host']", Qt.PartiallyChecked),
                  ColumnDisplayDefinition("v3['path']", Qt.Unchecked),
                  ColumnDisplayDefinition("paramstring(v3)", Qt.Unchecked),
                  ColumnDisplayDefinition("v4", Qt.PartiallyChecked),
                  ColumnDisplayDefinition("size(v5)", Qt.PartiallyChecked),
                  ColumnDisplayDefinition("cookies(v2)", Qt.Checked),
                  ColumnDisplayDefinition("setcookie(v5)", Qt.Checked)
                  ]
    activeColumns = []
    
    def __init__(self, connection, parent): 
        QAbstractTableModel.__init__(self, parent) 
        self.query = DBQuerying.Querier(connection, self)
        self.initColumns()
        self.loadData()
        
    def initColumns(self):
        self.activeColumns = []
        for transform in self.transforms:
            if transform.enabled: # is enabled?
                self.activeColumns.append(transform)

    def setTransforms(self, transforms):
        self.transforms = transforms
        self.initColumns()
        
    def setVariables(self, variables):
        self.fields = variables
    
    def onQueryChanged(self):
        self.reload()
        #TODO! tell the view to resize columns to contents, somehow
        #self.mongoTableView.resizeColumnsToContents()
        
        #super.resizeColumnsToContents()

    def getColumnDefinitions(self):
        """Returns the current transforms and colorcoding in tuples"""
        return self.transforms
    def getTransform(self, i):
        """Returns a specific transform"""
        return self.activeColumns[i].expression
    def getColumnTitle(self, i):
        """Returns a the title in the column definition for column i"""
        return self.activeColumns[i].title

    def getColumnColourSelection(self, i):
        """Returns the users choice of coloring for a specific column"""
        return self.activeColumns[i].coloring
        
    def getVariables(self):
        return self.fields
    
    def reload(self):
        self.modelAboutToBeReset.emit()
        self.loadData()
        self.modelReset.emit()
    
    def getData(self, item, fieldlist):
        """ This method reaches into a data entry according to dot-separated keys (fieldlist). 
        Example : item = {'url':{'parameter':'foo'}}, fieldlist ="url.parameter"
        ==> returns 'foo' 
        """
        key = fieldlist[0]
        data = None
        if item == None:
            return None
        if key in item:
            data = item[key]
        
        if len(fieldlist) == 1 : return data

        return self.getData(data, fieldlist[1:])

    def getDataArrayFromResult(self, item):
        """
        This method extracts the variable items from a database result. 
        Example : fields : ['method', 'url','url.path']
        ==> {'v1':'GET', 'v2':'http://foobar.com/foo', 'v3':'/foo'}
        """
        variables = {}
        for x in range(len(self.fields)):
            field = self.fields[x]
            key = "v%d" % x
            f = field.expression.split(".")
            value = self.getData(item, f)
            variables[key] = value
        return variables
    
    def loadData(self):
        """ Loads data from the database, and applies transformations to the loaded data. Also populates a 
        shadow-list containing the real mongodb-id:s , one for each row.  
        """
        #The data that will be shown in the table
        data = []
        #We need to keep track of the id's, so we can display the objects
        #in the object inspector
        object_ids = []
        
        rowCount = 0
        
        db_fields = []
        for field in self.fields:
            db_fields.append(field.expression)
        
        
        for item in self.query.find(fields=db_fields):
            rowData = []
            
            # Turn the db data into 'variables' array
            variables = self.getDataArrayFromResult(item)

            # Now we have the loaded data as variables v1, v2 etc in variables. 
            # Now map the columns by applying transforms
            for columnDef in self.activeColumns:
                if columnDef.enabled:
                    transformation = columnDef.expression
                    value = transformer.apply(transformation, variables)
                    rowData.append(str(value))
                
            #Add the transformed values as a new row
            data.append(rowData)
            #Keep track of each row's original object, i.e the id
            object_ids.append(item['_id'])
            rowCount += 1
            
        self.rc = rowCount
        self._data = data
        self._object_ids = object_ids
        
    def loadItem(self, id):
        """ Loads a single object from the database, does not apply any transformations.
        @todo this method should be a generic db connector method.  
        @param id the mongo id of the object 
        """
        return self.query.loadById(id)

    def getItemId(self, qIndex):
        """ Returns the mongo id for the underlying object which is shown at the specified qtableindex
        @param qindex the qtableindex of an item that is visible
        """
        row = qIndex.row()
        if row <= len(self._object_ids) : 
            return self._object_ids[row]
        print("Error, read out of bounds, objects id size: %d, specified row: %d" % (len(self._object_ids, row)))

    def rowCount(self, parent):
        """Returns the current number of rows"""
        return self.rc
    def columnCount(self, parent):
        """Returns the number of columns used"""
        return len(self.activeColumns)

    def data(self, index, role): 
        
        if not index.isValid(): 
            return QVariant()
#        print("%d, %d got visited by role %d"  % ( index.row(), index.column(), role))
        if index.row() < len(self._data):
            row = self._data[index.row()]
        else:
            return QVariant()
        
        if index.column() > len(row):
            return QVariant()
            
        if role == Qt.ToolTipRole or role == Qt.StatusTipRole: 
            return QVariant(row[index.column()])
        
        if role != Qt.DisplayRole and role != Qt.DecorationRole:
            return QVariant()
        
        #Check what we should show, either
        # 0) Text and no color
        # 1) Text with colour icon
        # 2) Only colour icon
        coloring = self.getColumnColourSelection(index.column())
        if coloring != 0 and role == Qt.DecorationRole:
            #The data to be rendered as a decoration (usually an icon).
            if len(self._data) > index.row():
                row = self._data[index.row()]
                (r, g, b) = getUniqueColor(row[index.column()])
                return QVariant(QColor(r, g, b))
        if coloring != 2 and role == Qt.DisplayRole: 
            if len(self._data) > index.row():
                row = self._data[index.row()]
                return QVariant(row[index.column()])
                
        return QVariant()
         
    def headerData(self, index, orientation, role):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if self.transforms == None: 
                    return QVariant("- all -") 
                return QVariant(self.getColumnTitle(index))
            return QVariant("row %d" % index)
    
    def sort(self, Ncol, order):
        """Sort table by given column number.
        """
        self.layoutAboutToBeChanged.emit()
        self._data = sorted(self._data, key=operator.itemgetter(Ncol))        
        if order == Qt.DescendingOrder:
            self._data.reverse()
        self.layoutChanged.emit()
